//Chris Hunter
/******************UNIMAC*****************
 This exercise begins to introduce some
 control over the medium. The server is
 effectively the transmit 1/2 of a full
 carrier-sensing MAC. It carrier-senses 
 and then transmits to each user,
 retransmitting when a user fails to ACK
 the original transmission. So, the users
 are responsible for the receive half of the
 full MAC layer. They must filter the
 destination addresses of received packets
 and ACK received DATA packets to prevent
 retransmissions.
 
 Note: The "uni" in "uniMac" only means that
 you are controlling the medium in only one
 direction. User code is still responsible
 for transmitting received Ethernet traffic
 to the server. This transmission, however,
 will not have the benefit of carrier sensing
 and retransmissions.
 *****************************************/

#include "warpmac.h"
#include "warpphy.h"
#include "xparameters.h"
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include "xutil.h"
#include "xcache_l.h"
#include "xexception_l.h"
#include "xcache_l.h"
#include "xutil.h"
#include "xexception_l.h"

#define ARPREQUEST 0x806

//Packet Types
#define DATAPACKET 0
#define ACKPACKET 1

//Routing table with agreed upon mapping between dipswitches and physical addresses
typedef struct {
	unsigned char addr[6];
} route;

//Byte array containing physical MAC address of this node
unsigned char myAddr[6];
//Instance of the routing table
route routeTable[16];
//Index to the routing table that identifies this node
unsigned char myID;


//Buffer for holding a packet-to-xmit across multiple retransmissions
Macframe txBuffer; 
//Buffer to hold received packet
Macframe rxBuffer;


int timerExpire(unsigned char timerType){
	/*This function is responsible for handling TIMEOUTs and BACKOFFs.
	It is registered using the warpmac_setTimerHandler function in
	main. The job responsibilities of this function are to:
	-increase the contention window upon the expiration of a TIMEOUT
	-initiate a BACKOFF timer upon the expiration of a TIMEOUT
	-retransmit a packet upon the expiration of a BACKOFF*/
	int status;
	switch(timerType){
		case TIMEOUT:
			if(txBuffer.isNew){
				status = warpmac_incrementResend(&txBuffer);
				if(status == 0){
					warpmac_enableEthernetInterrupt();
					return 0;
				}
				warpmac_setTimer(BACKOFF);
				return 0;
			}
			break;
		case BACKOFF:
			if(warpmac_carrierSense()){
				warpmac_sendOfdm(&txBuffer);
				warpmac_setTimer(TIMEOUT);
			}
			else{
				warpmac_setTimer(BACKOFF);
			}
			break;
			return 0;
	}
}

int ethernet_callback(Xuint32 length, char* payload){
	
	unsigned short int* pkttype = &payload[12];
	unsigned char* destNode;
	unsigned int* destIP;
	
	if(*pkttype==ARPREQUEST){
		destIP = &payload[38];
		destNode=&payload[41];
	}
	else{
		destIP = &payload[30];
		destNode=&payload[33];
	}
	
	if(*destIP>0xc0a80100 && *destIP<0xc0a8010F){
		warpmac_disableEthernetInterrupt();
		txBuffer.header.currReSend = 0;
		txBuffer.isNew = 1;
		txBuffer.header.length = length;
		txBuffer.header.pktType = DATAPACKET;
	
		//Set the modulation scheme for the packet's full-rate symbols
		txBuffer.header.fullRate = QPSK;
	
		//Copy in the packet's destination MAC address
		//Hard-coded for now as this node's partner node
		memcpy(txBuffer.header.destAddr,routeTable[*destNode].addr,6);
	
	
		if(warpmac_carrierSense()){
			warpmac_sendOfdm(&txBuffer);
			warpmac_setTimer(TIMEOUT);
		}
		else{
			warpmac_setTimer(BACKOFF);
		}
	}
		return 0;
}



int receiveBadPacket(Macframe* packet){
	/*This function blinks LEDs to show rate of bad packets.*/
	warpmac_incrementLEDLow();
}

int receiveGoodPacket(Macframe* packet) {
	warpmac_incrementLEDHigh();
	
	//IMPORTANT NOTE: Even though this function is passed a Macframe structure,
	//the data field is not yet filled. The user's MAC should process the packet
	//based on the header information, send out an ACK if necessary, and then
	//copy the payload out of the PHY with the warpmac_writePacket command.
	
	if(warpmac_addressedToMe(packet)){
		Macframe ackPacket;
		switch(packet->header.pktType){
			case ACKPACKET:
				if(warpmac_inTimeout()){
					warpmac_clearTimer(TIMEOUT);
					txBuffer.header.currReSend = 0;
					txBuffer.isNew = 0;
					warpmac_enableEthernetInterrupt();
				}
				break;
			case DATAPACKET:
				
				packet->header.currReSend = 0;
				packet->isNew = 1;
				warpmac_phyInterruptClear();
				warpmac_sendEthernet(packet);
				packet->header.currReSend = 0;
				packet->isNew = 0;
				
			break;

		}
	}
return 0;

}


int main(){
	//Initialize the framework
	unsigned char antSel = 0;
	
	//Read Dip Switch value from FPGA board.
	//This value will be used as an index into the routing table for other nodes
	myID = warpmac_getMyId();
	
	//Create an arbitrary address for this node
	unsigned char tmpAddr[6] = {0x16,0x24,0x63,0x53,0xe2,0xc2+myID};
	
	memcpy(myAddr,tmpAddr,6);
	
	//Fill an arbitrary routing table so that nodes know each others' addresses
	unsigned char i;
	for(i=0;i<16;i++){
		routeTable[i].addr[0] = myAddr[0];
		routeTable[i].addr[1] = myAddr[1];
		routeTable[i].addr[2] = myAddr[2];
		routeTable[i].addr[3] = myAddr[3];
		routeTable[i].addr[4] = myAddr[4];
		routeTable[i].addr[5] = myAddr[5]+i-myID;
	}
	
	//Initialize the framework
	warpmac_init();
	
	warpmac_setMacAddr(&myAddr);
	warpmac_setMaxResend(8);
	warpmac_setMaxCW(5);
	warpmac_setTimeout(160);
	warpmac_setSlotTime(9);
	
	warpmac_setRxBuffer(&rxBuffer,0);
	warpmac_setTxBuffer(1);
	
 	//Copy MAC address into source field
	memcpy(txBuffer.header.srcAddr,myAddr,6);
	
	warpmac_setGoodPacketHandler(receiveGoodPacket);
	warpmac_setBadPacketHandler(receiveBadPacket);
	
	warpmac_setTimerHandler(timerExpire);
	warpmac_setEmacHandler(ethernet_callback);

	//Set the RF center frequency
	xil_printf("Channel = %d\r\n",warpmac_setChannel(GHZ_2,8));
	
	warpmac_enableCSMA();
	warpmac_enableEthernetInterrupt();
	
	//Set the modulation scheme use for base rate (header) symbols
	// Every node must use the same mod scheme for base rate symbols!
	warpmac_setBaseRate(QPSK);
	while(1){
	}
}




